"""Module to generate OpenAPI documents and resources."""

from __future__ import annotations

import dataclasses
import fondat.annotation
import fondat.codec
import fondat.http
import fondat.resource
import fondat.types
import functools
import http
import inspect
import keyword
import types
import typing

from collections.abc import Iterable, Mapping
from datetime import date, datetime
from decimal import Decimal
from fondat.data import datacls
from fondat.security import Policy
from fondat.types import is_instance, is_optional, is_subclass
from types import NoneType
from typing import Annotated, Any, Literal, TypedDict, TypeVar
from uuid import UUID


class Default:
    def __init__(self, value: Any):
        self.value = value


class ComponentSchema:
    """
    Annotation to request schema be stored in the OpenAPI component schema section.

    Parameter:
    • name: component schema name
    """

    name = None  # can use class if name not required

    def __init__(self, name=None):
        self.name = name


_to_affix = []


def _affix(wrapped):
    _to_affix.append(wrapped)
    return wrapped


def _datacls(wrapped):
    return _affix(Annotated[datacls(wrapped), ComponentSchema])


Reference = TypedDict("Reference", {"$ref": str})
_affix(Reference)


@_datacls
class OpenAPI:
    openapi: str
    info: Info
    servers: Iterable[Server] | None
    paths: Paths
    components: Components | None
    security: Iterable[SecurityRequirement] | None
    tags: Iterable[Tag] | None
    externalDocs: ExternalDocumentation | None


@_datacls
class Info:
    title: str
    description: str | None
    termsOfService: str | None
    contact: Contact | None
    license: License | None
    version: str


@_datacls
class Contact:
    name: str | None
    url: str | None
    email: str | None


@_datacls
class License:
    name: str
    url: str | None


@_datacls
class Server:
    url: str
    description: str | None
    variables: Mapping[str, ServerVariable]


@_datacls
class ServerVariable:
    enum: Iterable[str] | None
    default: str = ""
    description: str | None = ""


@_datacls
class Components:
    schemas: Mapping[str, Schema | Reference] | None
    responses: Mapping[str, Response | Reference] | None
    parameters: Mapping[str, Parameter | Reference] | None
    examples: Mapping[str, Example | Reference] | None
    requestBodies: Mapping[str, RequestBody | Reference] | None
    headers: Mapping[str, Header | Reference] | None
    securitySchemes: Mapping[str, SecurityScheme | Reference] | None
    links: Mapping[str, Link | Reference] | None
    callbacks: Mapping[str, Callback | Reference] | None


@_datacls
class PathItem:
    summary: str | None
    description: str | None
    get: Operation | None
    put: Operation | None
    post: Operation | None
    delete: Operation | None
    options: Operation | None
    head: Operation | None
    patch: Operation | None
    trace: Operation | None
    servers: Iterable[Server] | None
    parameters: Iterable[Parameter | Reference] | None


Paths = Mapping[str, PathItem | Reference]
_affix(Paths)


@_datacls
class Operation:
    tags: Iterable[str] | None
    summary: str | None
    description: str | None
    externalDocs: ExternalDocumentation | None
    operationId: str | None
    parameters: Iterable[Parameter | Reference] | None
    requestBody: RequestBody | Reference | None
    responses: Responses
    callbacks: Mapping[str, Callback | Reference] | None
    deprecated: bool | None
    security: Iterable[SecurityRequirement] | None
    servers: Iterable[Server] | None


@_datacls
class ExternalDocumentation:
    description: str | None
    url: str


@_datacls
class Parameter:
    name: str
    in_: Literal["query", "header", "path", "cookie"]
    description: str | None
    required: bool | None
    deprecated: bool | None
    allowEmptyValue: bool | None
    style: str | None
    explode: bool | None
    allowReserved: bool | None
    schema: Schema | Reference | None
    example: Any | None
    examples: Mapping[str, Example | Reference] | None
    content: Mapping[str, MediaType] | None


@_datacls
class RequestBody:
    description: str | None
    content: Mapping[str, MediaType]
    required: bool | None


@_datacls
class MediaType:
    schema: Schema | Reference | None
    example: Any | None
    examples: Mapping[str, Example | Reference] | None
    encoding: Mapping[str, Encoding] | None


@_datacls
class Encoding:
    contentType: str | None
    headers: Mapping[str, Header | Reference] | None
    style: str | None
    explode: bool | None
    allowReserved: bool | None


@_datacls
class Response:
    description: str
    headers: Mapping[str, Header | Reference] | None
    content: Mapping[str, MediaType] | None
    links: Mapping[str, Link | Reference] | None


Responses = Mapping[str, Response | Reference]
_affix(Responses)


Callback = Mapping[str, PathItem]
_affix(Callback)


@_datacls
class Example:
    summary: str | None
    description: str | None
    value: Any | None
    externalValue: str | None


@_datacls
class Link:
    operationRef: str | None
    operationId: str | None
    parameters: Mapping[str, Any] | None
    requestBody: Any | None
    description: str | None
    server: Server | None


@_datacls
class Header:
    description: str | None
    required: bool | None
    deprecated: bool | None
    allowEmptyValue: bool | None
    style: str | None
    explode: bool | None
    allowReserved: bool | None
    schema: Schema | Reference | None
    example: Any | None
    examples: Mapping[str, Example | Reference] | None
    content: Mapping[str, MediaType] | None


@_datacls
class Tag:
    name: str
    description: str | None
    externalDocs: ExternalDocumentation | None


@_datacls
class Schema:
    title: str | None
    multipleOf: int | float | None
    maximum: int | float | None
    exclusiveMaximum: int | float | None
    minimum: int | float | None
    exclusiveMinimum: int | float | None
    maxLength: int | None
    minLength: int | None
    pattern: str | None
    maxItems: int | None
    minItems: int | None
    uniqueItems: bool | None
    maxProperties: int | None
    minProperties: int | None
    required: Iterable[str] | None
    enum: Iterable[Any] | None
    type: str | None
    allOf: Iterable[Schema | Reference] | None
    oneOf: Iterable[Schema | Reference] | None
    anyOf: Iterable[Schema | Reference] | None
    not_: Schema | Reference | None
    items: Schema | Reference | None
    properties: Mapping[str, Schema | Reference] | None
    additionalProperties: bool | Schema | Reference | None
    description: str | None
    format: str | None
    default: Any | None
    nullable: bool | None
    discriminator: Discriminator | None
    readOnly: bool | None
    writeOnly: bool | None
    xml: XML | None
    externalDocs: ExternalDocumentation | None
    example: Any | None
    deprecated: bool | None


@_datacls
class Discriminator:
    propertyName: str
    mapping: Mapping[str, str] | None


@_datacls
class XML:
    name: str | None
    namespace: str | None
    prefix: str | None
    attribute: bool | None
    wrapped: bool | None


@_datacls
class SecurityScheme:
    type: str
    description: str | None
    name: str | None
    in_: str | None
    scheme: str | None
    bearerFormat: str | None
    flows: OAuthFlows | None
    openIdConnectUrl: str | None


@_datacls
class OAuthFlows:
    implicit: OAuthFlow | None
    password: OAuthFlow | None
    clientCredentials: OAuthFlow | None
    authorizationCode: OAuthFlow | None


@_datacls
class OAuthFlow:
    authorizationUrl: str | None
    tokenUrl: str | None
    refreshUrl: str | None
    scopes: Mapping[str, str]


SecurityRequirement = Mapping[str, Iterable[str]]
_affix(SecurityRequirement)


# OpenAPI document graph complete; affix all type hints to avoid overhead
for dc in _to_affix:
    fondat.types.affix_type_hints(dc)


def _kwargs(python_type, annotations):
    kwargs = {}
    for annotation in annotations:
        match annotation:
            case str():
                kwargs["description"] = annotation
            case fondat.annotation.Description():
                kwargs["description"] = annotation.value
            case fondat.annotation.Example():
                with fondat.validation.validation_error_path("example"):
                    fondat.validation.validate(annotation.value, python_type)
                kwargs["example"] = fondat.codec.JSONCodec.get(python_type).encode(
                    annotation.value
                )
            case Default():
                with fondat.validation.validation_error_path("default"):
                    fondat.validation.validate(annotation.value, python_type)
                kwargs["default"] = fondat.codec.JSONCodec.get(python_type).encode(
                    annotation.value
                )
            case fondat.annotation.Deprecated():
                kwargs["deprecated"] = annotation.value
            case fondat.annotation.Deprecated:
                kwargs["deprecated"] = True
            case fondat.annotation.ReadOnly():
                kwargs["readOnly"] = annotation.value
            case fondat.annotation.ReadOnly:
                kwargs["readOnly"] = True
    return kwargs


providers = []


def _provider(wrapped=None):
    if wrapped is None:
        return functools.partial(_provider)
    providers.append(wrapped)
    return wrapped


def _simple_schema(pytype, schema_type, schema_format=None):
    @_provider
    def simple(*, python_type, annotations, **_):
        if python_type is pytype:
            return Schema(
                type=schema_type, format=schema_format, **_kwargs(python_type, annotations)
            )


_simple_schema(bool, "boolean")
_simple_schema(Decimal, "string", "decimal")
_simple_schema(datetime, "string", "date-time")
_simple_schema(date, "string", "date")  # must be after datetime
_simple_schema(UUID, "string", "uuid")


@_provider
def _str_schema(*, python_type, annotations, **_):
    if is_subclass(python_type, str):
        kwargs = {}
        for annotation in annotations:
            match annotation:
                case fondat.annotation.Format():
                    kwargs["format"] = annotation.value
                case fondat.validation.MinLen():
                    kwargs["minLength"] = annotation.value
                case fondat.validation.MaxLen():
                    kwargs["maxLength"] = annotation.value
                case fondat.validation.Pattern():
                    kwargs["pattern"] = annotation.pattern.pattern
        return Schema(type="string", **_kwargs(python_type, annotations), **kwargs)


@_provider
def _bytes_schema(*, python_type, annotations, **_):
    if is_subclass(python_type, (bytes, bytearray)):
        kwargs = {}
        for annotation in annotations:
            match annotation:
                case fondat.validation.MinLen():
                    kwargs["minLength"] = annotation.value
                case fondat.validation.MaxLen():
                    kwargs["maxLength"] = annotation.value
        return Schema(
            type="string",
            format="binary" if fondat.http.InBody in annotations else "byte",
            **_kwargs(python_type, annotations),
            **kwargs,
        )


@_provider
def _int_schema(*, python_type, annotations, **_):
    if is_subclass(python_type, int) and not is_subclass(python_type, bool):
        kwargs = {}
        for annotation in annotations:
            match annotation:
                case fondat.validation.MinValue():
                    kwargs["minimum"] = annotation.value
                case fondat.validation.MaxValue():
                    kwargs["maximum"] = annotation.value
        return Schema(
            type="integer", format="int64", **_kwargs(python_type, annotations), **kwargs
        )


@_provider
def _float_schema(*, python_type, annotations, **_):
    if is_subclass(python_type, float):
        kwargs = {}
        for annotation in annotations:
            match annotation:
                case fondat.validation.MinValue():
                    kwargs["minimum"] = annotation.value
                case fondat.validation.MaxValue():
                    kwargs["maximum"] = annotations.value
        return Schema(
            type="number", format="double", **_kwargs(python_type, annotations), **kwargs
        )


def _get_component_schema(annotations):
    for annotation in annotations:
        if annotation is ComponentSchema or is_instance(annotation, ComponentSchema):
            return annotation


@_provider
def _typeddict_schema(*, python_type, annotations, origin, args, processor, **_):
    if is_subclass(python_type, dict) and hasattr(python_type, "__annotations__"):
        if ref := processor.references.get(python_type):
            return ref
        component_schema = _get_component_schema(annotations)
        if component_schema:
            name = component_schema.name or processor.component_schema_name(
                python_type.__name__
            )
            ref = {"$ref": f"#/components/schemas/{name}"}
            processor.references[python_type] = ref
        hints = typing.get_type_hints(python_type, include_extras=True)
        required = list(python_type.__required_keys__) or None
        schema = Schema(
            type="object",
            properties={key: processor.schema(pytype) for key, pytype in hints.items()},
            required=required,
            additionalProperties=False,
            **_kwargs(python_type, annotations),
        )
        if component_schema:
            processor.openapi.components.schemas[name] = schema
            return ref
        return schema


@_provider
def _mapping_schema(*, python_type, annotations, origin, args, processor, **_):
    if is_subclass(origin, Mapping) and len(args) == 2:
        if args[0] is not str:
            raise TypeError("Mapping[k, v] only supports str keys")
        return Schema(
            type="object",
            properties={},
            additionalProperties=processor.schema(args[1]),
            **_kwargs(python_type, annotations),
        )


@_provider
def _iterable_schema(*, python_type, annotations, origin, args, processor, **_):
    if is_subclass(origin, Iterable) and not is_subclass(origin, Mapping) and len(args) == 1:
        kwargs = {}
        if is_subclass(origin, set):
            kwargs["uniqueItems"] = True
        for annotation in annotations:
            match annotation:
                case fondat.validation.MinLen():
                    kwargs["minItems"] = annotation.value
                case fondat.validation.MaxLen():
                    kwargs["maxItems"] = annotation.value
        return Schema(
            type="array",
            items=processor.schema(args[0]),
            **_kwargs(python_type, annotations),
            **kwargs,
        )


# keywords have _ suffix in dataclass fields (e.g. "in_", "for_", ...)
_dc_kw = {k + "_": k for k in keyword.kwlist}


@_provider
def _typevar_schema(*, python_type, processor, **_):
    if isinstance(python_type, TypeVar):
        return processor.schema(fondat.types.resolve_typevar(python_type))


@_provider
def _dataclass_schema(*, python_type, annotations, origin, processor, **_):
    if dataclasses.is_dataclass(python_type) or (origin and dataclasses.is_dataclass(origin)):
        if ref := processor.references.get(python_type):
            return ref
        component_schema = _get_component_schema(annotations)
        dc_type = origin if origin else python_type
        if component_schema:
            name = component_schema.name or processor.component_schema_name(dc_type.__name__)
            ref = {"$ref": f"#/components/schemas/{name}"}
            processor.references[python_type] = ref
        hints = typing.get_type_hints(dc_type, include_extras=True)
        required = {
            f.name
            for f in dataclasses.fields(dc_type)
            if f.default is dataclasses.MISSING
            and f.default_factory is dataclasses.MISSING
            and not is_optional(hints[f.name])
        }
        with fondat.types.capture_typevars(python_type):
            properties = {
                _dc_kw.get(key, key): processor.schema(pytype) for key, pytype in hints.items()
            }
        for key, schema in properties.items():
            if key not in required and not fondat.validation.is_valid(schema, Reference):
                schema.nullable = None
        schema = Schema(
            type="object",
            properties=properties,
            required=required or None,
            additionalProperties=False,
            **_kwargs(python_type, annotations),
        )
        if component_schema:
            processor.openapi.components.schemas[name] = schema
            return ref
        return schema


@_provider
def _union_schema(*, python_type, annotations, origin, args, processor, **_):
    if origin in {types.UnionType, typing.Union}:
        nullable = NoneType in args
        schemas = [processor.schema(arg) for arg in args if arg is not NoneType]
        if len(schemas) == 1:  # optional
            schema = schemas[0]
            if not fondat.validation.is_valid(schema, Reference):
                schema.nullable = True
            return schema
        return Schema(anyOf=schemas, nullable=nullable, **_kwargs(python_type, annotations))


@_provider
def _literal_schema(*, python_type, annotations, origin, args, processor, **_):
    if origin is Literal:
        nullable = None in args
        types = tuple({type(literal) for literal in args if literal is not None})
        enums = {t: [l for l in args if type(l) is t] for t in types}
        schemas = {t: processor.schema(t) for t in types}
        for t, s in schemas.items():
            if fondat.validation.is_valid(s, Reference):
                raise TypeError(f"Cannot document literals containing complex type: {t}")
            s.enum = enums[t]
        if len(types) == 1:  # homegeneous
            schema = schemas[types[0]]
            if nullable and not fondat.validate.is_valid(schema, Reference):
                schema.nullable = True
        else:
            schema = Schema(  # heterogeneus
                anyOf=list(schemas.values()),
                nullable=nullable,
                **_kwargs(python_type, annotations),
            )
        return schema


@_provider
def _any_schema(*, python_type, annotations, origin, args, **_):
    if python_type is Any:
        return Schema(**_kwargs(python_type, annotations))


class Processor:
    """Processes resource and populates OpenAPI document."""

    def __init__(self, openapi: OpenAPI):
        self.openapi = openapi
        self.references = {}
        self.schemes = {}

    def process(self, resource, path, params={}):
        if path == "/":
            path = ""
        tag = resource._fondat_resource.tag
        path_item = PathItem(
            parameters=[
                Parameter(
                    name=key,
                    in_="path",
                    required=True,
                    schema=self.schema(hint),
                )
                for key, hint in params.items()
            ]
            or None
        )
        for name in (n for n in dir(resource) if not n.startswith("_")):
            attr = getattr(resource, name)
            if res := self.resource(attr):
                self.process(
                    res,
                    f"{path}/{name}",
                    params,
                )
            elif operation := self.operation(tag, attr):
                fondat_op = getattr(attr, "_fondat_operation")
                if name == fondat_op.method:
                    setattr(path_item, name, operation)
                    self.openapi.paths[path or "/"] = path_item
                else:
                    op_item = PathItem(parameters=path_item.parameters)
                    setattr(op_item, fondat_op.method, operation)
                    self.openapi.paths[f"{path or ''}/{name}"] = op_item
        attr = getattr(resource, "__getitem__", None)
        if res := self.resource(attr):
            param_name, param_type = next(iter(typing.get_type_hints(attr).items()))
            if param_name in params:
                param_name = f"{res.__name__.casefold()}_{param_name}"
            while param_name in params:
                param_name = f"{param_name}_"
            self.process(
                res,
                f"{path}/{{{param_name}}}",
                {**params, param_name: param_type},
            )
        if self.schemes:
            if not self.openapi.components:
                self.openapi.components = Components()
            if not self.openapi.components.securitySchemes:
                self.openapi.components.securitySchemes = {}
            for name, scheme in self.schemes.items():
                if name not in self.openapi.components.securitySchemes:
                    match scheme:
                        case fondat.http.BasicScheme():
                            security_scheme = SecurityScheme(
                                type="http",
                                description=scheme.description,
                                scheme="basic",
                            )
                        case fondat.http.BearerScheme():
                            security_scheme = SecurityScheme(
                                type="http",
                                description=scheme.description,
                                scheme="bearer",
                                bearerFormat=scheme.format,
                            )
                        case fondat.http.HeaderScheme():
                            security_scheme = SecurityScheme(
                                type="apiKey",
                                description=scheme.description,
                                name=scheme.header,
                                in_="header",
                            )
                        case fondat.http.CookieScheme():
                            security_scheme = SecurityScheme(
                                type="apiKey",
                                description=scheme.description,
                                name=scheme.cookie,
                                in_="cookie",
                            )
                        case _:
                            raise TypeError(f"unknown scheme type for: {name}")
                    self.openapi.components.securitySchemes[name] = security_scheme

    @staticmethod
    def resource(obj):
        if fondat.resource.is_resource(obj):  # resource class or instance
            return obj
        if is_instance(obj, property):  # unbound property
            obj = obj.fget
        if callable(obj):
            try:
                returns = typing.get_type_hints(obj)["return"]
            except:
                return None
            if fondat.resource.is_resource(returns):
                return returns
        return None

    def operation(self, tag, method):
        if not callable(method):
            return None
        fondat_op = getattr(method, "_fondat_operation", None)
        if not fondat_op or not fondat_op.publish:
            return None
        op = Operation(parameters=[], responses={})
        op.tags = [tag]
        if fondat_op.summary:
            op.summary = fondat_op.summary
        if fondat_op.description:
            op.description = fondat_op.description
        if fondat_op.deprecated:
            op.deprecated = True
        hints = typing.get_type_hints(method, include_extras=True)
        parameters = inspect.signature(method).parameters
        for name, hint in hints.items():
            python_type, annotations = fondat.types.split_annotations(hint)
            if name == "return":
                if python_type is NoneType:
                    op.responses[str(http.HTTPStatus.NO_CONTENT.value)] = Response(
                        description="No content.",
                    )
                else:
                    origin = typing.get_origin(python_type)
                    args = typing.get_args(python_type)
                    if origin in {types.UnionType, typing.Union} and NoneType in args:
                        op.responses[str(http.HTTPStatus.NO_CONTENT.value)] = Response(
                            description="No content.",
                        )
                        python_type = fondat.types.union_type(
                            arg for arg in args if arg is not None
                        )
                        hint = (
                            Annotated[tuple([python_type, *annotations])]
                            if annotations
                            else python_type
                        )
                    op.responses[str(http.HTTPStatus.OK.value)] = Response(
                        description=self.description(annotations) or "Response.",
                        content={
                            fondat.codec.BinaryCodec.get(hint).content_type: MediaType(
                                schema=self.schema(hint)
                            )
                        },
                    )
            else:
                param = parameters[name]
                if param.default is not param.empty:
                    hint = Annotated[hint, Default(param.default)]
                param_in = fondat.http.get_param_in(method, name, hint)
                if is_instance(param_in, fondat.http.InQuery):
                    name = param_in.name
                    in_ = "query"
                    style = "form"
                    explode = False
                else:
                    continue  # ignore AsBody, InBody
                op.parameters.append(
                    Parameter(
                        name=name,
                        in_=in_,
                        description=self.description(annotations),
                        required=param.default is param.empty,
                        schema=self.schema(hint),
                        style=style,
                        explode=explode,
                    )
                )
        body_type = fondat.http.get_body_type(method)
        if body_type:
            python_type, annotations = fondat.types.split_annotations(body_type)
            op.requestBody = RequestBody(
                description=self.description(annotations),
                content={
                    fondat.codec.BinaryCodec.get(body_type).content_type: MediaType(
                        schema=self.schema(body_type)
                    )
                },
                required=True,
            )
        if "return" not in hints:
            op.responses[str(http.HTTPStatus.NO_CONTENT.value)] = Response(
                description="No content."
            )
        op.responses = {key: op.responses[key] for key in sorted(op.responses.keys())}
        if not op.parameters:
            op.parameters = None
        op.security = self.security_requirements(fondat_op)
        return op

    def security_requirements(self, operation):
        result = []
        for policy in (p for p in operation.policies or () if p.schemes is not None):
            requirements = {}
            for scheme in policy.schemes:
                if self.schemes.get(scheme.name) not in (None, scheme):
                    raise TypeError(f"different schemes with same name: {scheme.name}")
                self.schemes[scheme.name] = scheme
                requirements[scheme.name] = []  # no scopes in currently supported schemes
            result.append(requirements)
        return result if result else None

    @staticmethod
    def description(annotations):
        for annotation in annotations:
            match annotation:
                case str():
                    return annotation
                case fondat.annotation.Description():
                    return annotation.value

    def schema(self, type_hint, default=None):
        python_type, annotations = fondat.types.split_annotations(type_hint)
        origin = typing.get_origin(python_type)
        args = typing.get_args(python_type)
        for provider in providers:
            if (
                schema := provider(
                    python_type=python_type,
                    annotations=annotations,
                    origin=origin,
                    args=args,
                    processor=self,
                )
            ) is not None:
                return schema
        raise TypeError(f"failed to generate JSON Schema for type: {python_type}")

    def component_schema_name(self, name):
        if self.openapi.components is None:
            self.openapi.components = Components()
        if self.openapi.components.schemas is None:
            self.openapi.components.schemas = {}
        while name in self.openapi.components.schemas:
            name = f"{name}_"
        self.openapi.components.schemas[name] = "__reserved__"
        return name


def generate_openapi(*, resource: type, path: str = "/", info: Info) -> OpenAPI:
    """
    Generate an OpenAPI document for a resource.

    Parameters:
    • resource: resource to generate OpenAPI document for
    • path: URI path to resource
    • info: metadata about the API
    """
    openapi = OpenAPI(openapi="3.0.3", info=info, paths={})
    Processor(openapi).process(resource=resource, path=path)
    if openapi.paths:
        openapi.paths = {
            k: openapi.paths[k] for k in sorted(openapi.paths.keys(), key=str.lower)
        }
    if openapi.components and openapi.components.schemas:
        openapi.components.schemas = {
            k: openapi.components.schemas[k]
            for k in sorted(openapi.components.schemas.keys(), key=str.lower)
        }
    return openapi


def openapi_resource(
    *,
    resource: type,
    path: str = "/",
    info: Info,
    policies: Iterable[Policy] | None = None,
    publish: bool = False,
) -> Any:
    """
    Generate a resource that exposes an OpenAPI document for a given resource.

    Parameters:
    • resource: resource to generate OpenAPI document for
    • path: URI path to resource
    • info: provides metadata about the API
    • policies: security policies to apply to all operations
    • publish: publish the resource in documentation
    """

    @fondat.resource.resource
    class OpenAPIResource:
        def __init__(self):
            self.openapi = None

        @fondat.resource.operation(publish=publish, policies=policies)
        async def get(self) -> OpenAPI:
            if not self.openapi:
                self.openapi = generate_openapi(resource=resource, path=path, info=info)
            return self.openapi

    return OpenAPIResource()



"""Module to manage data in a SQL database."""

from __future__ import annotations

import builtins
import fondat.error
import fondat.patch
import fondat.security
import fondat.types
import hashlib
import logging
import re
import typing

from collections.abc import AsyncIterator, Iterable, Mapping
from contextlib import AbstractAsyncContextManager, suppress
from dataclasses import dataclass, is_dataclass
from fondat.codec import BinaryCodec, DecodeError, JSONCodec
from fondat.data import datacls
from fondat.error import BadRequestError, NotFoundError
from fondat.memory import MemoryResource
from fondat.pagination import Page, PaginationError
from fondat.patch import json_merge_patch
from fondat.resource import operation, query, resource
from fondat.types import is_optional
from fondat.validation import MinValue, ValidationError, validate
from functools import partial
from typing import Annotated, Any, Protocol, TypedDict, TypeVar


_logger = logging.getLogger(__name__)


Item = TypeVar("Item")
Row = TypeVar("Row")


@dataclass
class Param:
    """
    A parameterized value to include in an expression.

    Attributes:
    • value: the value of the parameter
    • type: the Python type of the parameter

    If the type of the value is not specified, then it is inferred by inspecting the
    value's type.
    """

    __slots__ = {"value", "type"}

    def __init__(self, value: Any, type: Any = None):
        self.value = value
        self.type = type if type else builtins.type(value)

    def __repr__(self) -> str:
        return f"Param(value={self.value}, type={self.type})"

    def __str__(self) -> str:
        return f"«{self.value}»"


Fragment = str | Param


class Expression(Iterable[Fragment]):
    """
    Represents a SQL expression as a sequence of fragments.

    Each fragment is one of: a string or Param object.

    An expression can be initialized with multiple arguments, each argument being any of:
    a string, a Param, an Expression, or an Iterable of any of the above.

    An expression can be extended (+=) with any of: a string, a Param, an Expression,
    or an Iterable of any of the above.
    """

    slots = {"fragments"}

    def __init__(self, *args):
        super().__init__()
        self.fragments = []
        for arg in args or ():
            self += arg

    def __repr__(self):
        return f"Expression({self.fragments!r})"

    def __str__(self):
        return "".join(str(f) for f in self.fragments)

    def __iter__(self):
        """Iterate over fragments of the expression."""
        return iter(self.fragments)

    def __bool__(self):
        """Return True if expression contains any fragments."""
        return bool(self.fragments)

    def __len__(self):
        """Return number of fragments in expression."""
        return len(self.fragments)

    def __iadd__(self, value):
        """Add fragment(s) to the expression."""
        match value:
            case str() | Param():
                self.fragments.append(value)
            case Iterable():
                for element in value:
                    self += element  # recursive
            case _:
                raise ValueError
        return self

    def __getitem__(self, key):
        return self.fragments[key]

    @staticmethod
    def join(
        value: Iterable[Expression | Fragment],
        sep: str | None = None,
    ) -> Expression:
        """Join expressions and/or fragments, separated by an optional string."""
        expr = Expression()
        for item in value:
            if expr and sep:
                expr += sep
            expr += item
        return expr


def _to_identifier(value: str):
    if value.isidentifier():
        return value
    return re.sub(r"[^A-Za-z_]", "_", value)


class Database(Protocol):
    """Base class for a SQL database."""

    async def execute(
        self,
        statement: Expression | str,
        result: type[Row] | None = None,
    ) -> AsyncIterator[Row] | None:
        """
        Execute a SQL statement.

        Parameter:
        • statement: SQL statement to excute
        • result: the type to return a query result row

        If the statement is a query that generates results, each row can be returned in
        a dataclass or dict object, whose type is specifed in the `result` parameter.
        Rows are provided via a returned asynchronous iterator.

        Must be called within a database transaction context.
        """
        raise NotImplementedError

    async def transaction(self) -> AbstractAsyncContextManager:
        """
        Return an asynchronous context manager, which scopes a transaction in which
        statement(s) are executed. Upon exit of the context, if an exception was raised,
        changes will be rolled back; otherwise changes will be committed.
        """
        raise NotImplementedError

    def sql_type(self, python_type: Any) -> str:
        """Return the SQL type that corresponds with the specified Python type."""
        raise NotImplementedError


async def select(
    *,
    database: Database,
    columns: Iterable[tuple[Expression, str, Any]],
    from_: Expression | None = None,
    where: Expression | None = None,
    group_by: Expression | None = None,
    having: Expression | None = None,
    order_by: Expression | None = None,
    offset: int | None = None,
    limit: int | None = None,
) -> AsyncIterator[dict[str, Any]]:
    """DEPRECATED: USE `select_iterator` OR `select_page`."""

    cols = {}
    for column in columns:
        name = _to_identifier(column[1])
        while name in cols:
            name += "_"
        cols[name] = column
    stmt = Expression("SELECT ")
    exprs = []
    for k, v in cols.items():
        expr = Expression(v[0])
        if len(v[0]) != 1 or k != v[0].fragments[0]:
            expr += f" AS {k}"
        exprs.append(expr)
    stmt += Expression.join(exprs, ", ")
    if from_ is not None:
        stmt += Expression(" FROM ", from_)
    if where is not None:
        stmt += Expression(" WHERE ", where)
    if group_by is not None:
        stmt += Expression(" GROUP BY ", group_by)
    if having is not None:
        stmt += Expression(" HAVING ", having)
    if order_by is not None:
        stmt += Expression(" ORDER BY ", order_by)
    if limit:
        stmt += f" LIMIT {limit}"
    if offset:
        stmt += f" OFFSET {offset}"
    stmt += ";"
    rows = await database.execute(
        stmt, TypedDict("Columns", {k: v[2] for k, v in cols.items()})
    )
    async for row in rows:
        yield {cols[k][1]: v for k, v in row.items()}


async def select_iterator(
    *,
    database: Database,
    columns: Mapping[str, Expression],
    from_: Expression | None = None,
    where: Expression | None = None,
    group_by: Expression | None = None,
    having: Expression | None = None,
    order_by: Expression | None = None,
    offset: int | None = None,
    limit: int | None = None,
    row_type: type[Row],
) -> AsyncIterator[Row]:
    """
    Execute a select statement, returning results through aynchronous row iterator.

    Parameters:
    • columns: mapping of row column names to select expressions
    • from_: FROM expression containing tables to select and/or tables joined
    • where: WHERE condition expression; None to match all rows
    • group_by: GROUP BY expression
    • having: HAVING clause expression
    • order_by: ORDER BY expression
    • offset: number of rows to skip, or None to skip none
    • limit: limit the number of rows returned, or None to select all
    • row_type: type to return for each row

    The row type can be either a dataclass or a typed dictionary. The name of each
    value in the row must correlate with the name of a result column.

    Must be called within a database transaction context.
    """

    aliases = {}
    for name in columns:
        alias = name
        if not alias.isidentifier():
            alias = re.sub(r"[^A-Za-z_]", "_", alias)
        while alias in aliases.values():
            alias += "_"
        if alias != name:
            aliases[name] = alias

    stmt = Expression("SELECT ")

    exprs = []
    for name, expr in columns.items():
        name = aliases.get(name, name)
        exprs.append(
            expr if len(expr) == 1 and expr[0] == name else Expression(expr, " AS ", name)
        )
    stmt += Expression.join(exprs, ",")

    if from_ is not None:
        stmt += Expression(" FROM ", from_)
    if where is not None:
        stmt += Expression(" WHERE ", where)
    if group_by is not None:
        stmt += Expression(" GROUP BY ", group_by)
    if having is not None:
        stmt += Expression(" HAVING ", having)
    if order_by is not None:
        stmt += Expression(" ORDER BY ", order_by)
    if limit:
        stmt += f" LIMIT {limit}"
    if offset:
        stmt += f" OFFSET {offset}"

    stmt += ";"

    result_type = (
        row_type
        if not aliases
        else TypedDict(
            "Row",
            **{
                aliases.get(name, name): type for name, type in row_type.__annotations__.items()
            },
        )
    )

    async for row in await database.execute(stmt, result_type):
        if result_type is not row_type:
            row = row_type(**{name: row[aliases.get(name, name)] for name in columns})
        validate(row, row_type)
        yield row


async def select_page(
    *,
    database: Database,
    columns: Mapping[str, Expression],
    from_: Expression | None = None,
    where: Expression | None = None,
    order_by: Expression,
    limit: int = 1000,
    cursor: bytes | None = None,
    item_type: type[Item],
) -> Page[Item]:
    """
    Execute a select statement, paginating results.

    Parameters:
    • columns: mapping of item column names to select expressions
    • from_: FROM expression containing tables to select and/or tables joined
    • where: WHERE condition expression; None to match all rows
    • order_by: ORDER BY expression
    • limit: limit the number of rows returned, or None to select all
    • item_type: type to return for each item in the page

    The item type can be either a dataclass or a typed dictionary. The name of each value in
    the item must correlate with the name of a result column. In order for pagination to
    operate safely, each returned item must be unique.

    Must be called within a database transaction context.
    """

    cursor_codec = BinaryCodec.get(tuple[int, bytes])
    item_codec = BinaryCodec.get(item_type)

    def hash_item(item: Item) -> bytes:
        return hashlib.sha256(item_codec.encode(item)).digest()

    (offset, hash) = cursor_codec.decode(cursor) if cursor else (None, None)

    _select = partial(
        select_iterator,
        database=database,
        columns=columns,
        from_=from_,
        where=where,
        order_by=order_by,
        row_type=item_type,
    )

    items = [item async for item in _select(offset=offset, limit=limit + 1)]

    if not items or (hash and hash != hash_item(items[0])):  # page drift
        items = []
        found = False
        async for item in _select():  # iterate all
            if not found:
                h = hash_item(item)
                if hash == h:  # index found
                    found = True
            if found:
                items.append(item)
                if len(items) > limit:
                    break
        if not items:
            raise PaginationError("pagination index lost")

    if len(items) > limit:
        hash = hash_item(items[-1])
        del items[-1]
        offset = (offset or 0) + len(items)
        cursor = cursor_codec.encode((offset, hash))
    else:
        cursor = None

    return Page[item_type](items, cursor)


class Table:
    """
    Represents a table in a SQL database.

    Parameters and attributes:
    • name: name of database table
    • database: database where table is managed
    • schema: dataclass or TypedDict type representing the table schema
    • pk: column name of primary key

    Attributes:
    • columns: mapping of column names to ther associated types
    """

    __slots__ = {"name", "database", "schema", "columns", "pk"}

    def __init__(self, name: str, database: Database, schema: type, pk: str):
        self.name = name
        self.database = database
        schema = fondat.types.strip_annotations(schema)
        if not is_dataclass(schema):
            raise TypeError("table schema must be a dataclass")
        self.schema = schema
        self.columns = typing.get_type_hints(schema, include_extras=True)
        if pk not in self.columns:
            raise ValueError(f"primary key not in schema: {pk}")
        self.pk = pk

    def __repr__(self):
        return f"Table(name={self.name}, schema={self.schema}, pk={self.pk})"

    async def create(self, execute: bool = True) -> Expression:
        """
        Create table in database. Must be called within a database transaction context.
        """
        stmt = Expression(f"CREATE TABLE {self.name} (")
        columns = []
        for column_name, column_type in self.columns.items():
            column = [column_name, self.database.sql_type(column_type)]
            if column_name == self.pk:
                column.append("PRIMARY KEY")
            if not is_optional(column_type):
                column.append("NOT NULL")
            columns.append(" ".join(column))
        stmt += Expression(", ".join(columns), ");")
        if execute:
            await self.database.execute(stmt)
        return stmt

    async def drop(self, execute: bool = True) -> Expression:
        """
        Drop table from database. Must be called within a database transaction context.
        """
        stmt = Expression(f"DROP TABLE {self.name};")
        if execute:
            await self.database.execute(stmt)
        return stmt

    async def select(
        self,
        *,
        columns: Iterable[str] | str | None = None,
        where: Expression | None = None,
        order_by: Iterable[str] | str | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> AsyncIterator[dict[str, Any]]:
        """
        Execute a SQL statement select rows from the table.

        Parameters:
        • columns: name(s) of column(s) to return, or None for all columns
        • where: statement containing WHERE expression, or None to match all rows
        • order_by: names of columns to order rows by, or None to not order rows
        • limit: limit the number of rows returned, or None to not limit
        • offset: number of rows to skip, or None to skip none

        Returns an asynchronous iterable for rows in table that match the where expression.
        Each row item is a dictionary that maps column name to value.

        Must be called within a database transaction context.
        """

        if isinstance(columns, str):
            columns = columns.replace(",", " ").split()

        if order_by is not None and not isinstance(order_by, str):
            order_by = ", ".join(order_by)

        columns = TypedDict(
            "Columns",
            {column: self.columns[column] for column in columns} if columns else self.columns,
        )

        async for row in select(
            database=self.database,
            columns=[
                (Expression(key), key, type)
                for key, type in typing.get_type_hints(columns, include_extras=True).items()
            ],
            from_=Expression(self.name),
            where=where,
            order_by=Expression(order_by) if order_by is not None else None,
            limit=limit,
            offset=offset,
        ):
            yield row

    async def count(self, where: Expression | None = None) -> int:
        """
        Return the number of rows in the table that match an optional expression.
        Must be called within a database transaction context.

        Parameters:
        • where: expression to match; None to match all rows
        """

        stmt = Expression(f"SELECT COUNT(*) AS count FROM {self.name}")
        if where:
            stmt += Expression(" WHERE ", where)
        stmt += ";"
        result = await self.database.execute(stmt, TypedDict("Result", {"count": int}))
        return (await result.__anext__())["count"]

    async def insert(self, value: Any) -> None:
        """
        Insert table row. Must be called within a database transaction context.
        """
        stmt = Expression(
            f"INSERT INTO {self.name} (",
            ", ".join(self.columns),
            ") VALUES (",
            Expression.join(
                (
                    Param(getattr(value, name), python_type)
                    for name, python_type in self.columns.items()
                ),
                ", ",
            ),
            ");",
        )
        await self.database.execute(stmt)

    async def read(self, key: Any) -> Any:
        """
        Return a table row, or None if not found. Must be called within a database transaction
        context.
        """
        try:
            return self.schema(
                **await self.select(
                    where=Expression(f"{self.pk} = ", Param(key, self.columns[self.pk])),
                    limit=1,
                ).__anext__()
            )
        except StopAsyncIteration:
            return None

    async def update(self, value: Any) -> None:
        """
        Update table row. Must be called within a database transaction context.
        """
        await self.database.execute(
            Expression(
                f"UPDATE {self.name} SET ",
                Expression.join(
                    (
                        Expression(f"{name} = ", Param(getattr(value, name), python_type))
                        for name, python_type in self.columns.items()
                    ),
                    ", ",
                ),
                f" WHERE {self.pk} = ",
                Param(getattr(value, self.pk), self.columns[self.pk]),
                ";",
            )
        )

    async def delete(self, key: Any) -> None:
        """
        Delete table row. Must be called within a database transaction context.
        """
        await self.database.execute(
            Expression(
                f"DELETE FROM {self.name} WHERE {self.pk} = ",
                Param(key, self.columns[self.pk]),
                ";",
            )
        )


class Index:
    """
    Represents an index on a table in a SQL database.

    Parameters:
    • name: name of index
    • table: table that the index defined for
    • keys: index keys (typically column names with optional order)
    • unique: is index unique
    """

    __slots__ = {"name", "table", "keys", "unique"}

    def __init__(
        self,
        name: str,
        table: Table,
        keys: Iterable[str],
        unique: bool = False,
    ):
        self.name = name
        self.table = table
        self.keys = keys
        self.unique = unique

    def __repr__(self):
        return f"Index(name={self.name}, table={self.table}, keys={self.keys}, unique={self.unique})"

    async def create(self, execute: bool = True) -> Expression:
        """
        Create index in database. Must be called within a database transaction context.
        """
        stmt = Expression("CREATE ")
        if self.unique:
            stmt += "UNIQUE "
        stmt += Expression(
            f"INDEX {self.name} ON {self.table.name} (",
            ", ".join(self.keys),
            ");",
        )
        if execute:
            await self.table.database.execute(stmt)
        return stmt

    async def drop(self, execute: bool = True) -> Expression:
        """
        Drop index from database. Must be called within a database transaction context.
        """
        stmt = Expression(f"DROP INDEX {self.name};")
        if execute:
            await self.table.database.execute(stmt)
        return stmt


def row_resource_class(
    table: Table,
    cache_size: int = 0,
    cache_expire: int | float = 1,
) -> type:
    """
    Return a base class for a row resource.

    Parameters:
    • table: table for which row resource is based
    • cache_size: number of rows to cache (evict least recently cached)
    • cache_expire: expire time for cached values in seconds
    """

    pk_type = table.columns[table.pk]

    cache = (
        MemoryResource(
            key_type=pk_type,
            value_type=table.schema,
            size=cache_size,
            evict=True,
            expire=cache_expire,
        )
        if cache_size
        else None
    )

    @resource
    class RowResource:
        """Row resource."""

        def __init__(self, pk: pk_type):
            self.table = table
            self.pk = pk

        async def _validate(self, value: table.schema):
            """Validate value; raise ValidationError if invalid."""
            validate(value, table.schema)

        async def _read(self):
            if cache:
                with suppress(NotFoundError):
                    return await cache[self.pk].get()
            row = await table.read(self.pk)
            if not row:
                raise NotFoundError
            if cache:
                await cache[self.pk].put(row)
            return row

        async def _insert(self, value: table.schema):
            if getattr(value, table.pk) != self.pk:
                raise ValidationError("primary key mismatch")
            await table.insert(value)
            if cache:
                await cache[self.pk].put(value)

        async def _update(self, old: table.schema, new: table.schema):
            if getattr(new, table.pk) != self.pk:
                raise ValidationError("primary key modified")
            if old != new:
                stmt = Expression(f"UPDATE {table.name} SET ")
                updates = []
                for name, python_type in table.columns.items():
                    ofield = getattr(old, name)
                    nfield = getattr(new, name)
                    if ofield != nfield:
                        updates.append(Expression(f"{name} = ", Param(nfield, python_type)))
                stmt += Expression.join(updates, ", ")
                stmt += Expression(f" WHERE {table.pk} = ", Param(self.pk, pk_type), ";")
                await table.database.execute(stmt)
            if cache:
                await cache[self.pk].put(new)

        @operation
        async def get(self) -> table.schema:
            """Get row from table."""
            if cache:
                with suppress(NotFoundError):
                    return await cache[self.pk].get()
            async with table.database.transaction():
                row = await table.read(self.pk)
            if not row:
                raise NotFoundError
            if cache:
                await cache[self.pk].put(row)
            return row

        @operation
        async def put(self, value: table.schema):
            """Insert or update (upsert) row."""
            await self._validate(value)
            async with table.database.transaction():
                try:
                    old = await self._read()
                    await self._update(old, value)
                except NotFoundError:
                    await self._insert(value)

        @operation
        async def patch(self, body: dict[str, Any]):
            """Modify row. Patch body is a JSON Merge Patch document."""
            async with table.database.transaction():
                old = await self._read()
                try:
                    new = json_merge_patch(value=old, type=table.schema, patch=body)
                except DecodeError as de:
                    raise BadRequestError from de
                await self._validate(new)
                await self._update(old, new)

        @operation
        async def delete(self) -> None:
            """Delete row."""
            if cache:
                with suppress(NotFoundError):
                    await cache[self.pk].delete()
            async with table.database.transaction():
                await table.delete(self.pk)

        @query
        async def exists(self) -> bool:
            """Return if row exists."""
            if cache:
                with suppress(NotFoundError):
                    await cache[self.pk].get()
                    return True
            where = Expression(f"{table.pk} = ", Param(self.pk, table.columns[table.pk]))
            async with table.database.transaction():
                return await table.count(where) != 0

    fondat.types.affix_type_hints(RowResource, localns=locals())
    return RowResource


def table_resource_class(table: Table, row_resource_type: type | None = None) -> type:
    """
    Return a base class for a table resource.

    Parameters:
    • table: table for which table resource is based
    • row_resource_type: type to instantiate for table row resource  [implict]
    """

    if row_resource_type is None:
        row_resource_type = row_resource_class(table)

    @datacls
    class Page:
        items: list[table.schema]
        cursor: bytes | None = None

    fondat.types.affix_type_hints(Page, localns=locals())

    dc_codec = JSONCodec.get(table.schema)
    pk_type = table.columns[table.pk]
    pk_codec = JSONCodec.get(pk_type)
    cursor_codec = BinaryCodec.get(pk_type)

    class TableResource:
        """Table resource."""

        def __getitem__(self, pk: pk_type) -> row_resource_type:
            return row_resource_type(pk)

        def __init__(self):
            self.table = table

        @operation
        async def get(
            self,
            limit: Annotated[int, MinValue(1)] = 1000,
            cursor: bytes | None = None,
        ) -> Page:
            """Get paginated list of rows, ordered by primary key."""
            if cursor is not None:
                where = Expression(
                    f"{table.pk} > ", Param(cursor_codec.decode(cursor), pk_type)
                )
            else:
                where = None
            async with table.database.transaction():
                items = [
                    table.schema(**result)
                    async for result in table.select(
                        order_by=table.pk, limit=limit, where=where
                    )
                ]
                cursor = (
                    cursor_codec.encode(getattr(items[-1], table.pk))
                    if len(items) == limit
                    else None
                )
            return Page(items=items, cursor=cursor)

        @operation
        async def patch(self, body: Iterable[dict[str, Any]]):
            """
            Insert and/or modify multiple rows in a single transaction.

            Patch body is an iterable of JSON Merge Patch documents; each document must
            contain the primary key of the row to patch.
            """
            async with table.database.transaction():
                for doc in body:
                    pk = doc.get(table.pk)
                    if pk is None:
                        raise ValidationError("missing primary key")
                    row = row_resource_type(pk_codec.decode(pk))
                    try:
                        old = await row._read()
                        try:
                            new = json_merge_patch(value=old, type=table.schema, patch=doc)
                        except DecodeError as de:
                            raise BadRequestError from de
                        await row._validate(new)
                        await row._update(old, new)
                    except NotFoundError:
                        new = dc_codec.decode(doc)
                        await row._validate(new)
                        await row._insert(new)

        @query
        async def find_pks(self, pks: set[table.columns[table.pk]]) -> list[table.schema]:
            """Return rows corresponding to the specified set of primary keys."""
            if not pks:
                return []
            async with table.database.transaction():
                return [
                    table.schema(**row)
                    async for row in table.select(
                        where=Expression(
                            f"{table.pk} IN (",
                            Expression.join(
                                (Param(pk, table.columns[table.pk]) for pk in pks), ", "
                            ),
                            ")",
                        )
                    )
                ]

    fondat.types.affix_type_hints(TableResource, localns=locals())
    return TableResource



"""Module to store resource items in memory."""

from __future__ import annotations

import fondat.error

from collections import namedtuple
from copy import deepcopy
from fondat.error import NotFoundError
from fondat.http import AsBody
from fondat.resource import mutation, operation, resource
from fondat.stream import Stream
from time import time
from typing import Annotated, Generic, TypeVar


K = TypeVar("K")
V = TypeVar("V")


_Item = namedtuple("Item", "value,time")


@resource
class MemoryResource(Generic[K, V]):
    """
    Represents a collection of items in memory.

    • key_type: type of item key
    • value_type: type of each item
    • size: maximum number of items to store  [unlimited]
    • evict: evict oldest item to make room for new item
    • expire: time to expire items in seconds  [unlimited]
    """

    def __init__(
        self,
        key_type: type[K],
        value_type: type[V],
        size: int | None = None,
        evict: bool = False,
        expire: int | float | None = None,
    ):
        if not getattr(key_type, "__hash__", None):
            raise TypeError("unhashable key_type: {key_type}")
        self.key_type = key_type
        if value_type is Stream:
            raise TypeError("value type not supported: {value_type}")
        self.value_type = value_type
        self.size = size
        self.evict = evict
        self.expire = expire
        self._storage: dict[K, _Item] = {}

    @operation
    async def get(self) -> set[K]:
        """Return collection item keys."""
        now = time()
        return {
            key
            for key, item in self._storage.items()
            if not self.expire or item.time + self.expire <= now
        }

    @mutation
    async def clear(self) -> None:
        """Remove all items from collection."""
        self._storage.clear()

    def __getitem__(self, key: K) -> "ItemResource[K, V]":
        return ItemResource(self, key)


@resource
class ItemResource(Generic[K, V]):
    """
    Represents an item in memory.

    Parameters:
    • memory: resource where item resides
    • key: item key in collection
    """

    def __init__(self, memory: MemoryResource, key: K):
        self.memory = memory
        self.key = key

    @operation
    async def get(self) -> V:
        """Get item."""
        item = self.memory._storage.get(self.key)
        if not item or (self.memory.expire and time() > item.time + self.memory.expire):
            raise NotFoundError
        return deepcopy(item.value)

    @operation
    async def put(self, value: Annotated[V, AsBody]) -> None:
        """Store item."""
        now = time()
        if self.memory.expire:  # purge expired entries
            for key in {
                key
                for key, item in self.memory._storage.items()
                if item.time + self.memory.expire <= now
            }:
                del self.memory._storage[key]
        if self.memory.size and self.memory.evict:  # evict oldest entry
            Oldest = namedtuple("Oldest", "key,time")
            self.memory._storage.pop(self.key, None)
            while len(self.memory._storage) >= self.memory.size:
                oldest = None
                for key, item in self.memory._storage.items():
                    if not oldest or item.time < oldest.time:
                        oldest = Oldest(key, item.time)
                if oldest:
                    self.memory._storage.pop(oldest.key, None)
        if self.memory.size and len(self.memory._storage) >= self.memory.size:
            raise fondat.error.errors.InsufficientStorageError
        self.memory._storage[self.key] = _Item(value, now)

    @operation
    async def delete(self):
        """Delete item."""
        await self.get()  # ensure item exists and has not expired
        self.memory._storage.pop(self.key, None)
